@@ -3,8 +3,8 @@ use strict; use warnings FATAL => 'all';
 use parent 'Exporter';
 use Carp qw(confess);
 
-use 5.018002;
-our $VERSION = '1.02';
+use 5.010;
+our $VERSION = '1.03';
 
 use Cwd;
 use DTL::Fast::Template;
@@ -12,7 +12,6 @@ use DTL::Fast::Template;
 our @EXPORT_OK;
 
 # @todo Texts should be tossed by references
-# @todo Protection from cycled ierarchy
 # @todo These should be done via Cache class which can work with last modified date
 our %TEMPLATES_CACHE = ();
 our $TEMPLATES_CACHE_HITS = 0;
@@ -52,10 +51,16 @@ sub get_template
     }    
     else
     {
-        $template = _read_template($template_name, $dirs);
+        my $template_path;
+
+        ($template, $template_path) = _read_template($template_name, $dirs);
         $template =~ s/\{\% (?:block|endblock|extends) .*?\%\}//gs;
         
-        my @arguments = ($template, $dirs);
+        my @arguments = (
+            $template
+            , $dirs
+            , 'file_path' => $template_path
+        );
         push @arguments, 'ssi_dirs', $kwargs{'ssi_dirs'}
             if $kwargs{'ssi_dirs'};
         push @arguments, 'url_source', $kwargs{'url_source'}
@@ -73,18 +78,30 @@ sub _apply_inheritance
 {
     my $template = shift;
     my $dirs = shift;
-
+    my $inheritance_path = shift;
+    
     if( $template =~ s/\s*\{\% extends\s*"(.+?)" \%\}//s ) # template has inheritance
     {
-        my $parent_template = _read_template($1, $dirs);
+        my $parent_name = $1;
+        my( $parent_template, $parent_template_path) = _read_template($parent_name, $dirs, $inheritance_path);
         
-        my %named_blocks = (
-            $template =~ /\{\% block ([^\s]+) \%\}(.+?)\{\% endblock \%\}/gs
-        );
-        
-        my $block_names = join '|', keys(%named_blocks);
-        $parent_template =~ s/\{\% block ($block_names) \%\}.+?\{\% endblock \%\}/\{\% block $1 \%\}$named_blocks{$1}\{\% endblock \%\}/gsi;
-        $template = $parent_template;
+        if( defined $parent_template )
+        {
+            my %named_blocks = (
+                $template =~ /\{\% block ([^\s]+) \%\}(.+?)\{\% endblock \%\}/gs
+            );
+            
+            my $block_names = join '|', keys(%named_blocks);
+            $parent_template =~ s/\{\% block ($block_names) \%\}.+?\{\% endblock \%\}/\{\% block $1 \%\}$named_blocks{$1}\{\% endblock \%\}/gsi;
+            $template = $parent_template;
+        }
+        else
+        {
+            die sprintf( "Couldn't found a parent template: %s in one of the following directories: %s"
+                , $parent_name
+                , join( ', ', @$dirs)
+            );
+        }
     }
     
     return $template;
@@ -94,8 +111,10 @@ sub _read_template
 {
     my $template_name = shift;
     my $dirs = shift;
+    my $inheritance_path = shift // { '*path' => [] };
 
     my $template = undef;
+    my $template_path = undef;
 
     my $cache_key = sprintf '%s:%s:%s'
         , __PACKAGE__
@@ -109,9 +128,26 @@ sub _read_template
     }
     else
     {
-        $template = _read_file($template_name, $dirs);
-        $template = _apply_inheritance($template, $dirs)
-            if defined $template;
+        ($template, $template_path) = _read_file($template_name, $dirs);
+        
+        if( defined $template )
+        {
+            if( exists $inheritance_path->{$template_path} )
+            {
+                die sprintf(
+                    "Recursive inheritance detected:\n%s\n" 
+                    , join "\n inherited from ", @{$inheritance_path->{'*path'}}, $template_path
+                );
+            }
+            
+            push @{$inheritance_path->{'*path'}}, $template_path;
+            $inheritance_path->{$template_path} = 1;
+            
+            $template = _apply_inheritance($template, $dirs, $inheritance_path);
+                
+            pop @{$inheritance_path->{'*path'}};
+            delete $inheritance_path->{$template_path};
+        }
             
         $TEMPLATES_CACHE{$cache_key} = $template;
     }
@@ -121,7 +157,7 @@ Unable to find template %s in directories:
 %s
 _EOT_
     
-    return $template;
+    return ($template, $template_path);
 }
 
 sub _read_file
@@ -129,11 +165,12 @@ sub _read_file
     my $template_name = shift;
     my $dirs = shift;
     my $template;
-
+    my $template_path;
+    
     foreach my $dir (@$dirs)
     {
         $dir =~ s/[\/\\]+$//gsi;
-        my $template_path = sprintf '%s/%s', $dir, $template_name;
+        $template_path = sprintf '%s/%s', $dir, $template_name;
         if( 
             -e $template_path
             and -f $template_path
@@ -157,7 +194,7 @@ sub _read_file
         }        
     }
     
-    return $template;
+    return ($template, $template_path);
 }
 
 # result should be cached with full list of params
@@ -196,7 +233,7 @@ DTL::Fast - Perl implementation of Django templating language.
 
 =head1 VERSION
 
-Version 1.02
+Version 1.03
 
 =head1 SYNOPSIS
 
@@ -442,6 +479,7 @@ is being rendered like:
 =head1 BENCHMARKS
 
 I've compared module speed with previous abandoned implementation: L<C<Dotiac::DTL>> in both modes: FCGI and CGI. Test template and scripts are in /timethese directory.
+Django templating in Python with cache works about 80% slower than C<DTL::Fast>.
 
 =head2 FCGI/mod_perl
 
@@ -480,13 +518,33 @@ Tests shows, that C<DTL::Fast> works 26% slower, than L<C<Dotiac::DTL>> in CGI e
 
 =over
 
+=item * 13/01/2015 - v1.03
+
+=over
+
+=item * Tested with CentOS & Perl 5.10
+
+=item * Lowered Perl version requirement to 5.10
+
+=item * Changed implicit split to explicit in C<wordcount> filter (v5.10 considers it depricated).
+
+=item * Added exception on missing parent template in C<extends> tag.
+
+=item * Added exception on missing included template in C<include> tag.
+
+=item * Added exception on recursive inheritance (C<extends> tag).
+
+=item * Added exception on recursive inclusion (C<include> tag).
+
+=back
+
 =item * 10/01/2015 - v1.02
 
 =over
 
 =item * changed some intermediate getters to direct access. Improved rendering performance by 10%. 
 
-=item * added tests for performance measuring and profiling (see C</timethese> directory).
+=item * added tests for performance measuring and profiling (see C<timethese> directory).
 
 =back
 
